local _, playerClass = UnitClass("player")
local playerHealer = (playerClass == "PRIEST") or (playerClass == "SHAMAN") or (playerClass == "PALADIN") or (playerClass == "DRUID")
local playerCaster = (playerClass == "MAGE") or (playerClass == "PRIEST") or (playerClass == "WARLOCK")
local playerMelee = (playerClass == "ROGUE") or (playerClass == "WARRIOR") or (playerClass == "HUNTER")
local playerHybrid = (playerClass == "DEATHKNIGHT") or (playerClass == "DRUID") or (playerClass == "PALADIN") or (playerClass == "SHAMAN")

--Libraries
local L = AceLibrary("AceLocale-2.2"):new("DrDamage")
local GT = LibStub:GetLibrary("LibGratuity-3.0")
local DrDamage = DrDamage

--General
local settings
local _G = getfenv(0)
local type = type
local pairs = pairs
local ipairs = ipairs
local tonumber = tonumber
local math_floor = math.floor
local math_min = math.min
local math_max = math.max
local string_match = string.match
local string_format = string.format
local string_find = string.find
local string_sub = string.sub
local string_gsub = string.gsub
local string_len = string.len
local select = select

--Module
local UnitBuff = UnitBuff
local UnitDebuff = UnitDebuff
local UnitMana = UnitMana
local UnitLevel = UnitLevel
local UnitDamage = UnitDamage
local UnitPowerType = UnitPowerType
local BOOKTYPE_SPELL = BOOKTYPE_SPELL
local GetSpellName = GetSpellName
local GetSpellInfo = GetSpellInfo
local GetMacroSpell = GetMacroSpell
local GetActionInfo = GetActionInfo
local GetCursorInfo = GetCursorInfo
local GetInventoryItemLink = GetInventoryItemLink
local GetInventorySlotInfo = GetInventorySlotInfo
local GetItemInfo = GetItemInfo
local GetItemGem = GetItemGem
local GetTime = GetTime
local GetWeaponEnchantInfo = GetWeaponEnchantInfo
local HasAction = HasAction
local IsEquippedItem = IsEquippedItem
local SecureButton_GetModifiedAttribute = SecureButton_GetModifiedAttribute
local SecureButton_GetEffectiveButton = SecureButton_GetEffectiveButton
local ActionButton_GetPagedID = ActionButton_GetPagedID

--Module variables
local playerCompatible, loadedTalents
local spellInfo, talentInfo, talents
local updatingSpell, updateSetItems
local dmgMod
local DrD_Font = GameFontNormal:GetFont()
DrDamage.visualChange = true

--Local functions
function DrDamage.ClearTable(table)
	for k in pairs( table ) do
		table[k] = nil
	end
end
function DrDamage.Round(x, y)
	return math_floor( x * 10 ^ y + 0.5 ) / 10 ^ y
end
function DrDamage.MatchData( data, ... )
	if not data or not ... then
		return false
	end

	if type( data ) == "table" then
		for i = 1, select('#', ...) do
			if data[1] then
				for _, dataName in ipairs( data ) do
					for i = 1, select('#', ...) do
						if dataName == select(i, ...) then
							return true
						end
					end
				end
			else
				if data[select(i, ...)] then
					return true
				end
			end
		end
	else	
		for i = 1, select('#', ...) do
			if data == select(i, ...) then
				return true
			end
		end
	end
	
	return false
end

local DrD_ClearTable = DrDamage.ClearTable
local DrD_Round = DrDamage.Round
local DrD_MatchData = DrDamage.MatchData

local function DrD_Set( n, set, change )
	return function(v) 
		settings[n] = v
		if change then
			DrDamage.visualChange = true
		end
		if not set and not DrDamage:IsEventScheduled("DrD_FullUpdate") then
			DrDamage:CancelSpellUpdate()
			DrDamage:ScheduleEvent("DrD_FullUpdate", DrDamage.UpdateAB, 1.0, DrDamage)
		end
	end
end

local ABstop, ABdefault, ABdisable, ABfunc --, ABhidden
local ABtable, ABobjects, ABrefresh = {}, {}, {}

local function DrD_DetermineAB()
	--Default
	ABdefault = true
	for i = 1, 6 do
		for j = 1, 12 do
			table.insert(ABobjects,_G[((select(i,"ActionButton", "MultiBarBottomLeftButton", "MultiBarBottomRightButton", "MultiBarRightButton", "MultiBarLeftButton", "BonusActionButton"))..j)])
		end
	end
	
	if IsAddOnLoaded("Bartender4") then
		local func = function(button) if button.Secure then return button.Secure:GetActionID() else return button:GetActionID() end end
		ABrefresh["BT4Button"] = function()
			for i=1,120 do
				ABtable["BT4Button"..i] = func
			end
		end
		ABdefault = false
	end
	if IsAddOnLoaded("Nurfed") then
		local func = function(button)
			if button.spell then
				if button.type == "spell" then
					local pid = Nurfed:getspell(button.spell)
					if pid then
						return nil, GetSpellName(pid, BOOKTYPE_SPELL)
					end
				elseif button.type == "macro" then
					local action, rank = GetMacroSpell(button.spell)
					if action then 							
						return nil, action, rank
					end
				end
			end
		end	
		ABrefresh["Nurfed_Button"] = function()
			for i=1,120 do
				ABtable["Nurfed_Button"..i] = func
			end
		end
		ABdefault = false
	end
	if IsAddOnLoaded("Macaroon") then
		local func = function(button)
			if button.config and button.config.type == "action" then
				return SecureButton_GetModifiedAttribute(button,"action",SecureButton_GetEffectiveButton(button))
			end
			if button.macroshow then
				return nil, string_match(button.macroshow,"[^%(]+"), button.macrorank
			end
			if button.macrospell then
				return nil, string_match(button.macrospell,"[^%(]+"), button.macrorank
			end
		end
		ABrefresh["MacaroonButton"] = function()
			for _, button in pairs(Macaroon.Buttons) do
				ABtable[button[1]:GetName()] = func
			end
		end
	end	
	if IsAddOnLoaded("Dominos") then
		local func = function(button) return SecureButton_GetModifiedAttribute(button,"action",SecureButton_GetEffectiveButton(button)) end
		ABrefresh["DominosActionButton"] = function()
			for i=1,120 do
				ABtable["DominosActionButton"..i] = func
			end
		end
	end
	if IsAddOnLoaded("IPopBar") then
		local func = function(button) return SecureButton_GetModifiedAttribute(button,"action",SecureButton_GetEffectiveButton(button)) end
		ABrefresh["IPopBarButton"] = function()
			for i=1,120 do
				ABtable["IPopBarButton"..i] = func
			end
		end
	end	
	if IsAddOnLoaded( "FlexBar2" ) then
		ABstop = true
		ABfunc = function()
			for _, button in pairs(FlexBar2.Buttons) do
				button:UpdateTextSub("drd")
			end
		end
	end
	if nUI_DrDamageIntegration then
		ABrefresh["nUI"] = nUI_DrDamageIntegration( ABtable )
		ABdefault = false
	end	
	if IsAddOnLoaded("CT_BarMod") then
		--There's an object list available CT_BarMod.actionButtonList, index = actionID, value.hasAction
		local func = function(button) return button.object.id end
		ABrefresh["CT_BarModActionButton"] = function()
			for i=13,24 do
				ABtable["CT_BarModActionButton"..i] = func
			end
			for i=49,120 do
				ABtable["CT_BarModActionButton"..i] = func
			end			
		end
	end
	if IsAddOnLoaded("RDX") then
		local func = function(button) return button:GetAttribute("action") end
		ABrefresh["RDX_ActionBars"] = function()
			for i=1,120 do
				ABtable["VFLButton"..i] = func
			end
		end
	end
	
	if ABdefault then
		settings.BlizzardAB = true
	end
	
	DrDamage:RefreshAB()
end

function DrDamage:RefreshAB()
	if ABrefresh then
		for k in pairs(ABtable) do
			ABtable[k] = nil
		end
		for _, func in pairs(ABrefresh) do
			func()
		end
	end
end

--Defaults
local defaults = DrDamage.defaults
defaults.ABText = true
defaults.FontEffect = "OUTLINE"
defaults.FontSize = 10
defaults.FontXPosition = 0
defaults.FontYPosition = 0
defaults.FontColorDmg = { r = 1, g = 1, b = 0.2 }
defaults.FontColorHeal = { r = 0.4, g = 1.0, b = 0.3 }
defaults.UpdateShift = false
defaults.UpdateAlt = false
defaults.UpdateCtrl = false
defaults.BlizzardAB = false
defaults.AlwaysTooltip = true
defaults.CombatTooltip = false
defaults.NeverTooltip = false
defaults.AltTooltip = false
defaults.CtrlTooltip = false
defaults.ShiftTooltip = false
defaults.DefaultColor = false
defaults.SwapCalc = false
defaults.PlayerAura = {}
defaults.TargetAura = {}
defaults.Consumables = {}
DrDamage:RegisterDB("DrDamageDB");
DrDamage:RegisterDefaults("profile", defaults)
DrDamage.options = { type='group', args = {} }
DrDamage:RegisterChatCommand({ "/drdmg", "/drdamage" }, DrDamage.options)

--FuBar options
if AceLibrary:HasInstance("FuBarPlugin-2.0") then
	DrDamage.name = "DrDamage"
	DrDamage.title = "Dr. Damage"
	DrDamage.hasNoColor = true
	DrDamage.hasNoText = true
	DrDamage.cannotDetachTooltip = true
	DrDamage.hideWithoutStandby = true
	DrDamage.independentProfile = true
	DrDamage.blizzardTooltip = true
	DrDamage.hasIcon = playerMelee and "Interface\\Icons\\Ability_DualWield" or "Interface\\Icons\\Spell_Holy_SearingLightPriest"
	DrDamage.OnMenuRequest = DrDamage.options
end

function DrDamage:OnTooltipUpdate()
	GameTooltip:AddLine( "          Dr. Damage" )
	GameTooltip:AddLine( "Hint: Right-Click for options", 0, 1, 0, true )
end

function DrDamage:OnInitialize()
	settings = self.db.profile	
end

function DrDamage:OnEnable()
	if IsAddOnLoaded("FlexBar2") and FlexBar2:HasModule("TextSubs") then
		FlexBar2:GetModule("TextSubs"):RegisterTextSub("drd", 
		function(button)
			if not playerCompatible then return false end
			if not settings.ABText then return "" end
			local spell = button:GetModifiedAttribute("spell");
			if(spell and spell ~= "") then
				-- Run the spell through GetSpellName to seperate Name & Rank and get the rank if its not specified
				local name, rank = GetSpellName(spell)
				return self:CheckAction(nil, nil, nil, name, rank)
			end
		end)		
	end
	if self.PlayerData then
		self.ClassSpecials = {}
		self.DmgCalculation = {}
		self.SetBonuses = {}
		self.RelicSlot = {}
		self.talents = {}
		self:PlayerData()
		self.PlayerData = nil
		spellInfo = self.spellInfo
		talentInfo = self.talentInfo
		self:CommonData()
		self.CommonData = nil
		talents = self.talents
		--Make talent spell tables better suited for iteration
		for k in pairs( self.talentInfo ) do
			local talent = self.talentInfo[k]
			for i=1,#talent do
				local spelltable = talent[i].Spells
				if type(spelltable) == "table" then
					for j,v in ipairs( spelltable ) do
						spelltable[v] = true
						spelltable[j] = nil
					end
				end
			end
		end
	elseif not self.spellInfo and not self.talentInfo then
		return
	end
	
	self:OnProfileEnable()
	
	if self.GeneralOptions then
		self:GeneralOptions()
		self.GeneralOptions = nil
	end
  	if AceLibrary("AceEvent-2.0"):IsFullyInitialized() then
  		self:Load()
  	else
  		self:RegisterEvent("PLAYER_ENTERING_WORLD", "ScheduleLoad")
  	end
end

function DrDamage:CommonData()
	if (select(2, UnitRace("player")) == "Draenei") and (playerClass == "PALADIN" or playerClass == "PRIEST" or playerClass == "SHAMAN" or playerClass == "MAGE")  then
		--Gift of the Naaru
		spellInfo[GetSpellInfo(28880)] = {
					["Name"] = "Gift of the Naaru",
					[0] = { School = { "Holy", "Healing" }, Cooldown = 180, bonusFactor = 1, eDot = true, eDuration = 15, sTicks = 3 },
					[1] = { 35, 35, },
		}
		self.Calculation[GetSpellInfo(28880)] = function( calculation )
			local b, p, n = UnitAttackPower("player")
			local ap = b + p + n
			if ap > calculation.spellDmg then calculation.spellDmg = ap end
			calculation.minDam = calculation.minDam + UnitLevel("player") * 15
			calculation.maxDam = calculation.minDam
		end
	end
end	

function DrDamage:ScheduleLoad()
	self:UnregisterEvent("PLAYER_ENTERING_WORLD")
	self:ScheduleEvent("DrD_Load", self.Load, 2, self)
end

function DrDamage:Load()
	if self.Caster_OnEnable then
		self:Caster_OnEnable()
		self.Caster_OnEnable = nil
	end
	if self.Melee_OnEnable then
		self:Melee_OnEnable()
		self.Melee_OnEnable = nil
	end
	if DrD_DetermineAB then 
		DrD_DetermineAB()
		DrD_DetermineAB = nil
	end
	if self.Caster_InventoryChanged then
		self:Caster_InventoryChanged()
	end
	if self.Melee_InventoryChanged then
		self:Melee_InventoryChanged()
	end
	
	dmgMod = select(7, UnitDamage("player"))
	self:MetaGems(true)
	self:UpdateGlyphs()
	playerCompatible = true
	
	if not loadedTalents then
		--Create talent tables
		for k, v in pairs( spellInfo ) do
			v["Talents"] = {}
			if v["Secondary"] then
				v["Secondary"]["Talents"] = {}
			end
		end	
		self:UpdateTalents()
	else
		self:UpdateAB()
	end
	
	self:Hook(GameTooltip, "SetAction", true)
	self:SecureHook(GameTooltip, "SetSpell")  
	self:RegisterEvent("AceDB20_ResetDB")
 	self:RegisterEvent("UNIT_INVENTORY_CHANGED")
  	self:RegisterEvent("CHARACTER_POINTS_CHANGED","UPDATE_TALENTS")
  	self:RegisterEvent("PLAYER_TALENT_UPDATE","UPDATE_TALENTS")
	self:RegisterEvent("GLYPH_ADDED", "UpdateGlyphs")
	self:RegisterEvent("GLYPH_UPDATED", "UpdateGlyphs")
	self:RegisterEvent("GLYPH_REMOVED", "UpdateGlyphs")
	
	if settings.ABText then
		if settings.UpdateAlt or settings.UpdateCtrl or settings.UpdateShift then
			self:RegisterEvent( "MODIFIER_STATE_CHANGED" )
		end
		self:RegisterEvent("ACTIONBAR_PAGE_CHANGED")
		self:RegisterEvent("ACTIONBAR_HIDEGRID")
		self:RegisterEvent("EXECUTE_CHAT_LINE")
		self:RegisterEvent("UPDATE_SHAPESHIFT_FORM")
		self:RegisterEvent("PLAYER_TARGET_CHANGED")
		self:RegisterEvent("UPDATE_MACROS")
		self:RegisterEvent("LEARNED_SPELL_IN_TAB")
		self:RegisterEvent("UNIT_AURA")
	end
end

function DrDamage:OnDisable()
	if playerCompatible then
		local tempType = settings.ABText
		settings.ABText = false
		self:UpdateAB()
		settings.ABText = tempType
	end	
end

function DrDamage:OnProfileEnable()
	if self.Caster_OnProfileEnable then self:Caster_OnProfileEnable() end
	if self.Melee_OnProfileEnable then self:Melee_OnProfileEnable() end
	settings = self.db.profile
end

function DrDamage:GeneralOptions()
	self.options.args.General = {
		type = "group", 
		desc = L["General Options"], 
		name = L["General"],
		order = 1,
		args = {
			BlizzardAB = {
				type = 'toggle',
				name = L["Support for Blizzard AB even with other AB addons running."],
				desc = L["Toggles updating on for the default actionbar."],
				order = 10,
				get =  function() return settings["BlizzardAB"] end,
				set =  DrD_Set("BlizzardAB"),				
			},
			SwapCalc = {
				type = 'toggle',
				name = L["Swap calculation for double purpose abilities"],
				desc = L["Toggles between the two calculations for abilities."],
				order = 15,
				get = function() return settings["SwapCalc"] end,
				set = DrD_Set("SwapCalc", false, true),
			},			
			ABText = {
				type = 'toggle',
				name = L["Actionbar text on/off"],
				desc = L["Toggles the actionbar text on/off"],
				order = 20,
				get = function() return settings["ABText"] end,
				set = DrD_Set("ABText"),
			},		
			FontSize = {
				type = 'range',
				name = L["Actionbar font size"],
				desc = L["Set the font size"],
				min = 6,
				max = 20,
				step = 1,
				order = 30,
				get =  function() return settings["FontSize"] end,
				set = DrD_Set("FontSize", false, true),
			},
			FontEffect = {
				type = "text",
				name = L["Actionbar font effect"],
				desc = L["Set the font effect"],
				validate = { L["Outline"], L["ThickOutline"], L["None"] },
				order = 40,
				get =	function()
	
						if settings.FontEffect == "" then
							return L["None"]
						elseif settings.FontEffect == "THICKOUTLINE" then
							return L["ThickOutline"]
						else
							return L["Outline"]
						end
					end,
				set = 	function(v)
						if v == L["None"] then
							settings.FontEffect = ""
						elseif v == L["ThickOutline"] then
							settings.FontEffect = "THICKOUTLINE"
						else
							settings.FontEffect = "OUTLINE"
						end

						self.visualChange = true
						self:UpdateAB()
					end,
			},				
			FontXPosition = {
				type = 'range',
				name = L["Actionbar font horizontal position"],
				desc = L["Set the font horizontal position"],
				min = -5,
				max = 5,
				step = 1,
				order = 50,
				get =  function() return settings["FontXPosition"] end,
				set = 	DrD_Set("FontXPosition", false, true),
			},
			FontYPosition = {
				type = 'range',
				name = L["Actionbar font vertical position"],
				desc = L["Set the font vertical position"],
				min = -15,
				max = 30,
				step = 1,
				order = 60,
				get =  function() return settings["FontYPosition"] end,
				set = 	DrD_Set("FontYPosition", false, true),
			},
			FontColorDmg = {
				type = "text",
				name = L["Damage text color"],
				desc = L["Set the actionbar damage text color"],
				usage = L["<r=n, g=n, b=n> where n is 0-1"],
				order = 70,
				get = function()
					return "r="..settings.FontColorDmg.r..", g="..settings.FontColorDmg.g..", b="..settings.FontColorDmg.b
				end,
				set = function(colors)
					for k, v in string.gmatch(string.lower(colors), "([r|g|b])=([%d\.]+)") do
						if tonumber(v) >= 0 and tonumber(v) <= 1 then
							settings.FontColorDmg[k] = v
						end
					end
					self.visualChange = true
					self:UpdateAB()						
				end
			},
			Update = {
				type = 'group', 
				name = L["Update"],
				desc = L["Update"],
				order = 90,
				args = {
					Update = {
						name = L["Update"],
						desc = L["Forces update to actionbar"],
						type = "execute",
						order = 1,
						func = function() self.visualChange = true, self:UpdateAB() end,
					},
					UpdateShift = {
						type = 'toggle',
						name = L["Updates actionbars on shift modifier key."],

						desc = L["Toggles updates on modifier key on/off."],
						order = 2,
						get =  function() return settings["UpdateShift"] end,
						set =  function(v)
								settings["UpdateShift"] = v
								if v then
									if not self:IsEventRegistered("MODIFIER_STATE_CHANGED") then
										self:RegisterEvent("MODIFIER_STATE_CHANGED")
									end
								elseif not settings.UpdateAlt and not settings.UpdateCtrl then
									self:UnregisterEvent("MODIFIER_STATE_CHANGED")
								end
							end,				
					},
					UpdateAlt = {
						type = 'toggle',
						name = L["Updates actionbars on alt modifier key."],
						desc = L["Toggles updates on modifier key on/off."],
						order = 3,
						get =  function() return settings["UpdateAlt"] end,
						set =  function(v)
								settings["UpdateAlt"] = v
								if v then
									if not self:IsEventRegistered("MODIFIER_STATE_CHANGED") then
										self:RegisterEvent("MODIFIER_STATE_CHANGED")
									end
								elseif not settings.UpdateCtrl and not settings.UpdateShift then
									self:UnregisterEvent("MODIFIER_STATE_CHANGED")
								end
							end,				
					}, 
					UpdateCtrl = {
						type = 'toggle',
						name = L["Updates actionbars on ctrl modifier key."],
						desc = L["Toggles updates on modifier key on/off."],
						order = 4,
						get =  function() return settings["UpdateCtrl"] end,
						set =  function(v)
								settings["UpdateCtrl"] = v
								if v then
									if not self:IsEventRegistered("MODIFIER_STATE_CHANGED") then
										self:RegisterEvent("MODIFIER_STATE_CHANGED")
									end
								elseif not settings.UpdateAlt and not settings.UpdateShift then
									self:UnregisterEvent("MODIFIER_STATE_CHANGED")
								end
							end,				
					},
				},
			},
			DisplayTooltip = {
				type = 'group',
				name = L["Display tooltip"],
				desc = L["Control when the tooltip is displayed"],
				order = 100,
				args = {
					Always = {
						type = 'toggle',
						name = L["Always"],
						desc = L["Always display the tooltip"],
						order = 1,
						get = function() return settings["AlwaysTooltip"] end,
						set = function(v)
							settings.AlwaysTooltip = v
							settings.NeverTooltip = not v
							settings.ShiftTooltip = false
							settings.CtrlTooltip = false
							settings.AltTooltip = false
							if v then
								settings.CombatTooltip = false
							end
	
						end,
					},
					Combat = {
						type = 'toggle',
						name = "Disable in combat",
						desc = "Disables tooltip display in combat without modifier key. Set a modifier key if needed.",
						order = 3,
						get = function() return settings["CombatTooltip"] end,
						set = function(v)
							settings.CombatTooltip = v
							settings.NeverTooltip = false
							if v then
								settings.AlwaysTooltip = false
							elseif not settings.CtrlTooltip and not settings.ShiftTooltip and not settings.AltTooltip then
								settings.AlwaysTooltip = true
							end
						end,
					},					
					Never = {
						type = 'toggle',
						name = L["Never"],
						desc = L["Never display the tooltip"],
						order = 2,
						get = function() return settings["NeverTooltip"] end,
						set = function(v)
							settings.AlwaysTooltip = not v
							settings.NeverTooltip = v
							settings.CombatTooltip = false
							settings.ShiftTooltip = false
							settings.CtrlTooltip = false
							settings.AltTooltip = false
						end,
					},
					Alt = {
						type = 'toggle',
						name = L["With Alt"],
						desc = L["Display tooltip when Alt is pressed"],
						order = 4,
						get = function() return settings["AltTooltip"] end,
						set = function(v)
							settings.NeverTooltip = false
							settings.AltTooltip = v
							if v then
								settings.AlwaysTooltip = false
							elseif not settings.CtrlTooltip and not settings.ShiftTooltip and not settings.CombatTooltip then
								settings.AlwaysTooltip = true
							end
						end,
					},
					Ctrl = {
						type = 'toggle',
						name = L["With Ctrl"],
						desc = L["Display tooltip when Ctrl is pressed"],
						order = 5,
						get = function() return settings["CtrlTooltip"] end,
						set = function(v)
							settings.NeverTooltip = false
							settings.CtrlTooltip = v
							if v then
								settings.AlwaysTooltip = false
							elseif not settings.AltTooltip and not settings.ShiftTooltip and not settings.CombatTooltip then
								settings.AlwaysTooltip = true
							end
						end,
					},
					Shift = {
						type = 'toggle',
						name = L["With Shift"],
						desc = L["Display tooltip when Shift is pressed"],
						order = 6,
						get = function() return settings["ShiftTooltip"] end,
						set = function(v)
							settings.NeverTooltip = false
							settings.ShiftTooltip = v
							if v then
								settings.AlwaysTooltip = false
							elseif not settings.CtrlTooltip and not settings.AltTooltip and not settings.CombatTooltip then
								settings.AlwaysTooltip = true
							end
						end,
					},
					DefaultColor = {
						type = 'toggle',
						name = L["Default tooltip colors"],
						desc = L["Toggles the default blizzard tooltip colors on/off"],
						order = 10,
						get = function() return settings["DefaultColor"] end,
						set = DrD_Set("DefaultColor", true),
					},					
				},
			},
			Aura = {
				type = 'group',
				name = L["Modify Buffs/Debuffs"],
				desc = L["Choose what buffs/debuffs to always include into calculations. Abilities will mostly NOT double stack if buff/debuff is really applied."],
				order = 170,
				args = {
					Player = {
						type = 'group',
						name = L["Player"],
						desc = L["Choose player buffs/debuffs."],
						args = {},
						},
					Target = {
						type = 'group',
						name = L["Target"],
						desc = L["Choose target buffs/debuffs."],
						args = {},
						},
					Consumables = {
						type = 'group',
						name = L["Consumables"],
						desc = L["Choose consumables to include."],
						args = {},
						},
				},
			},			
		},
	}
	local optionsTable = self.options.args.General.args
	local auraTable = optionsTable.Aura.args.Player.args
	for k,v in pairs( self.PlayerAura ) do
		if not v.ModType or v.ModType ~= "Update" then
			auraTable[(string_gsub(k," +", ""))] = {
				type = 'toggle',
				name = k,
				desc = L["Include "] .. k,
				get = function() return settings["PlayerAura"][k] end,
				set = function(v) settings["PlayerAura"][k] = v or nil; self:UpdateAB() end,
			}
		end
	end
	auraTable = optionsTable.Aura.args.Target.args
	for k,v in pairs( self.TargetAura ) do
		if not v.ModType or v.ModType ~= "Update" then
			auraTable[(string_gsub(k," +", ""))] = {
				type = 'toggle',
				name = k,
				desc = L["Include "] .. k,
				get = function() return settings["TargetAura"][k] end,
				set = function(v) settings["TargetAura"][k] = v or nil; self:UpdateAB() end,
			}
		end
	end
	auraTable = optionsTable.Aura.args.Consumables.args
	for k,v in pairs( self.Consumables ) do
		auraTable[(string_gsub(k," +", ""))] = {
			type = 'toggle',
			name = k,
			desc = L["Include "] .. k,
			get = function() return settings["Consumables"][k] end,
			set = function(v) settings["Consumables"][k] = v or nil; self:UpdateAB() end,
		}
	end	
	if playerHealer then
		optionsTable.FontColorHeal = {
			type = "text",
			name = L["Heal text color"],
			desc = L["Set the actionbar heal text color"],
			usage = L["<r=n, g=n, b=n> where n is 0-1"],
			order = 80,
			get = function()
				return "r="..settings.FontColorHeal.r..", g="..settings.FontColorHeal.g..", b="..settings.FontColorHeal.b
			end,
			set = function(colors)
				for k, v in string.gmatch(string.lower(colors), "([r|g|b])=([%d\.]+)") do
					if tonumber(v) >= 0 and tonumber(v) <= 1 then
						settings.FontColorHeal[k] = v
					end
				end
				self.visualChange = true 
				self:UpdateAB()
			end
		}
	end
end

--EVENTS:
function DrDamage:CancelSpellUpdate()
	updatingSpell = nil
	self:CancelScheduledEvent("DrD_SpellUpdate")
end

--REQUIRES DELAY:
function DrDamage:ACTIONBAR_PAGE_CHANGED()
	self.visualChange = true
	if not self:IsEventScheduled("DrD_FullUpdate") then
		self:CancelSpellUpdate()
		self:ScheduleEvent("DrD_FullUpdate", self.UpdateAB, 0.2, self)
	end
end
function DrDamage:ACTIONBAR_HIDEGRID()
	self.visualChange = true
	if not self:IsEventScheduled("DrD_FullUpdate") and (GetCursorInfo() == "spell" or GetCursorInfo() == "macro") then
		self:CancelSpellUpdate()
		self:ScheduleEvent("DrD_FullUpdate", self.UpdateAB, 0.2, self)
	end
end
function DrDamage:UPDATE_SHAPESHIFT_FORM()
	self.visualChange = true
	if not self:IsEventScheduled("DrD_FullUpdate") then
		self:CancelSpellUpdate()
		self:ScheduleEvent("DrD_FullUpdate", self.UpdateAB, 0.75, self)
	end
end
function DrDamage:EXECUTE_CHAT_LINE()
	self:CancelScheduledEvent("DrD_StopMacroWatch")
	self:RegisterEvent("ACTIONBAR_SLOT_CHANGED")
	self:ScheduleEvent("DrD_StopMacroWatch", self.UnregisterEvent, 0.5, self, "ACTIONBAR_SLOT_CHANGED")
end
function DrDamage:ACTIONBAR_SLOT_CHANGED(id)
	local gtype, pid = GetActionInfo(id)
	if gtype == "macro" then
		local name = GetMacroSpell(pid)
		if name and spellInfo[name] then
			self.visualChange = true
			if not self:IsEventScheduled("DrD_FullUpdate") then
				self:ScheduleEvent("DrD_SlotUpdate", self.UpdateAB, 0.5, self, nil, id)
			end
		end
	end
end

--INSTANT FULL UPDATES
function DrDamage:AceDB20_ResetDB( name, dab )
	if dab == self.db.name then
		self.visualChange = true
		self:UpdateAB()
	end
end
function DrDamage:UNIT_INVENTORY_CHANGED( unit )
	if unit == "player" then
		if self.Caster_InventoryChanged and self:Caster_InventoryChanged() then
			updateSetItems = true
			self:CancelScheduledEvent("DrD_FullUpdate")
			self:UpdateAB()
			self:ScheduleEvent("DrDamage_UpdateMetaGem", self.MetaGems, 0.5, self)
			return
		end
		if self.Melee_InventoryChanged and self:Melee_InventoryChanged() then
			updateSetItems = true
			self:CancelScheduledEvent("DrD_FullUpdate")
			self:UpdateAB()
			self:ScheduleEvent("DrDamage_UpdateMetaGem", self.MetaGems, 0.5, self)
			return
		end
	end
end

DrDamage.Melee_critMBonus = 0
DrDamage.Spell_critMBonus = 0
DrDamage.Healing_critMBonus = 0

function DrDamage:MetaGems( initial )
	local oldactive, newactive
	
	if self.Melee_critMBonus ~= 0 or self.Spell_critMBonus ~= 0 or self.Healing_critMBonus ~= 0 then
		self.Melee_critMBonus = 0
		self.Spell_critMBonus = 0
		self.Healing_critMBonus = 0
		oldactive = true
	end

	local helm = GetInventoryItemLink("player", 1)
	if helm then
		for i = 1, 3 do 
			local mgem = tonumber(string_match(select(2, GetItemGem(helm, i)) or "",".-item:(%d+).*"))
			if mgem == 34220 or mgem == 41285 then
				if not self:IsMetaGemInactive() then
					self.Spell_critMBonus = 0.03
					newactive = true
				end
				break
			elseif mgem == 41376 then
				if not self:IsMetaGemInactive() then
					self.Healing_critMBonus = 0.03
					newactive = true
				end
				break
			elseif mgem == 32409 or mgem == 41398 then
				if not self:IsMetaGemInactive() then
					self.Melee_critMBonus = 0.03
					newactive = true
				end
				break
			end
		end
	end
	if not initial and newactive ~= oldactive then
		self:CancelScheduledEvent("DrD_FullUpdate")
		self:UpdateAB()
	end
end

function DrDamage:IsMetaGemInactive()
	GT:SetInventoryItem("player", 1)
	for j = 1, GT:NumLines() do 
		if GT:GetLine(j) and string_find(GT:GetLine(j), "|cff808080.*808080") then
			return true
		end
	end
end

--DELAY FOR EFFICIENCY
function DrDamage:PLAYER_TARGET_CHANGED()
	if not self:IsEventScheduled("DrD_FullUpdate") then
		self:CancelSpellUpdate()
		self:ScheduleEvent("DrD_FullUpdate", self.UpdateAB, 0.5, self)
	end
	self:TargetAuraUpdate(true)
end
function DrDamage:UPDATE_TALENTS()
	self.visualChange = true
	if not self:IsEventScheduled("DrD_TalentUpdate") then
		self:ScheduleEvent("DrD_TalentUpdate", self.UpdateTalents, 1, self)
	end
end
function DrDamage:UPDATE_MACROS()
	self.visualChange = true
	if not self:IsEventScheduled("DrD_FullUpdate") then
		self:CancelSpellUpdate()
		self:ScheduleEvent("DrD_FullUpdate", self.UpdateAB, 0.75, self)
	end
end
function DrDamage:LEARNED_SPELL_IN_TAB()
	if not self:IsEventScheduled("DrD_FullUpdate") then
		self:CancelSpellUpdate()
		self:ScheduleEvent("DrD_FullUpdate", self.UpdateAB, 1.5, self)
	end
end
function DrDamage:UNIT_AURA(unit)
	if unit == "player" then
		if not self:IsEventScheduled("DrD_PlayerAuraUpdate") then
			self:ScheduleEvent("DrD_PlayerAuraUpdate", self.PlayerAuraUpdate, 0.5, self)
		end
	elseif unit == "target" then
		if not self:IsEventScheduled("DrD_TargetAuraUpdate") then
			self:ScheduleEvent("DrD_TargetAuraUpdate", self.TargetAuraUpdate, 1, self)
		end	
	end
end

local PlayerAura = DrDamage.PlayerAura
local TargetAura = DrDamage.TargetAura
local oPlayerAura = {}
local nPlayerAura = {}
local oTargetAura = {}
local nTargetAura = {}

function DrDamage:PlayerAuraUpdate()
	if self:IsEventScheduled("DrD_FullUpdate") then
		return
	end	
	if dmgMod ~= select(7, UnitDamage("player")) then
		dmgMod = select(7, UnitDamage("player"))
		self:UpdateAB()
		return
	end
	if self.Caster_CheckBaseStats and self:Caster_CheckBaseStats() then
		self:UpdateAB()
		return
	end
	if self.Melee_CheckBaseStats and self:Melee_CheckBaseStats() then
		self:UpdateAB()
		return
	end
	--[[ Enable if ever needed
	if self.Calculation["PlayerAura"] then
		local update, spell = self.Calculation["PlayerAura"]
		if update then
			return self:UpdateAB(spell)
		end
	end
	--]]
	for i=1,40 do
		local name, rank, texture, count = UnitBuff("player",i)
		if name then
			if PlayerAura[name] then			
				nPlayerAura[name] = (rank or "") .. (count or "")
			end
		else
			break
		end
	end
	for i=1,40 do
		local name, rank, texture, count = UnitDebuff("player",i)
		if name then
			if PlayerAura[name] then
				nPlayerAura[name] = (rank or "") .. (count or "")
			end
		else 
			break
		end
	end
	local buffName, multi
	--Buff/debuff gained or count/rank changed
	for k,v in pairs(nPlayerAura) do
		if not oPlayerAura[k] then
			multi = buffName
			buffName = k
		elseif 	oPlayerAura[k] ~= v then
			multi = buffName
			buffName = k
		end
	end	
	--Buff/debuff lost
	for k,v in pairs(oPlayerAura) do
		if not nPlayerAura[k] then
			multi = buffName
			buffName = k
		end
		oPlayerAura[k] = nil
	end
	--Copy new table to old and clear
	for k,v in pairs(nPlayerAura) do
		oPlayerAura[k] = v
		nPlayerAura[k] = nil
	end
	if buffName then
		local updateSpell = not multi and PlayerAura[buffName].Spell

		if updatingSpell and updatingSpell == updateSpell then
			return
		end
		if updateSpell then
			updatingSpell = updateSpell
			self:ScheduleEvent("DrD_SpellUpdate", self.UpdateAB, 0.5, self, updateSpell)
		else
			self:CancelSpellUpdate()
			self:ScheduleEvent("DrD_FullUpdate", self.UpdateAB, 0.5, self)
		end
	end
end

function DrDamage:TargetAuraUpdate( changed )
	if self:IsEventScheduled("DrD_FullUpdate") and not changed then
		return
	end
	if self.Calculation["TargetAura"] then
		local update, spell = self.Calculation["TargetAura"]
		if update and not changed then
			return self:UpdateAB(spell)
		end
	end		
	if playerHealer then
		for i=1,40 do
			local name, rank, texture, count = UnitBuff("target",i)
			if name then
				if TargetAura[name] then			
					nTargetAura[name.."|"..texture] = (rank or "") .. (count or "")
				end
			else break end
		end
	end
	for i=1,40 do
		local name, rank, texture, count = UnitDebuff("target",i)
		if name then
			if TargetAura[name] then
				nTargetAura[name.."|"..texture] = (rank or "") .. (count or "")
			end
		else break end
	end
	local buffName, multi
	--Buff/debuff gained or count/rank changed
	for k,v in pairs(nTargetAura) do
		if not oTargetAura[k] then
			multi = buffName
			buffName = string_match(k,"[%w%s]+")
		elseif 	oTargetAura[k] ~= v then
			multi = buffName
			buffName = string_match(k,"[%w%s]+")
		end
	end	
	--Buff/debuff lost
	for k,v in pairs(oTargetAura) do
		if not nTargetAura[k] then
			multi = buffName
			buffName = string_match(k,"[%w%s]+")
		end
		oTargetAura[k] = nil
	end
	--Copy new table to old and clear
	for k,v in pairs(nTargetAura) do
		oTargetAura[k] = v
		nTargetAura[k] = nil
	end		
	if TargetAura[buffName] and not changed then
		local updateSpell = not multi and TargetAura[buffName].Spell

		if updatingSpell and updatingSpell == updateSpell then
			return
		end
		if updateSpell then
			updatingSpell = updateSpell
			self:ScheduleEvent("DrD_SpellUpdate", self.UpdateAB, 0.8, self, updateSpell )				
		else
			self:CancelSpellUpdate()
			self:ScheduleEvent("DrD_FullUpdate", self.UpdateAB, 1.0, self)
		end
	end
end

local lastState = GetTime()
function DrDamage:MODIFIER_STATE_CHANGED( state )
	if (state == "LALT" or state == "RALT") and settings.UpdateAlt or (state == "LCTRL" or state == "RCTRL") and settings.UpdateCtrl or (state == "LSHIFT" or state == "RSHIFT") and settings.UpdateShift then
		if GetTime() - lastState < 0.3 then
			if self:IsEventScheduled("DrD_FullUpdate_Modifier") then
				self:CancelScheduledEvent( "DrD_FullUpdate_Modifier" )
				self.visualChange = false
			end
		else
			if not self:IsEventScheduled("DrD_FullUpdate_Modifier") then
				self:ScheduleEvent("DrD_FullUpdate_Modifier", self.UpdateAB, 0.3, self)
			end
			self.visualChange = true
		end
		lastState = GetTime()
	end
end

local TalentIterate
function DrDamage:UpdateTalents(manual, all)
	if not manual then
		DrD_ClearTable( talents )
		local talentTable
		if not loadedTalents then
			self.options.args.General.args.Talents = { type = "group", name = L["Modify Talents"], desc = L["Modify talents manually. Modified talents are not saved between sessions."], order = 200, args = {} }
			talentTable = self.options.args.General.args.Talents.args
			talentTable.Reset = { type = "execute", name = L["Reset Talents"], desc = L["Reset talents to your current talent configuration."], order = 1, func = function() self:UpdateTalents() end }
			loadedTalents = true
		end

		for t = 1, GetNumTalentTabs() do	
			for i = 1, GetNumTalents(t) do
				local talentName, _, _, _, currRank, maxRank = GetTalentInfo(t, i)

				if talentTable and self.talentInfo[talentName] and not self.talentInfo[talentName].NoManual then
					talentTable[(string_gsub(talentName," +", ""))] = {
							type = 'range',
							name = talentName,
							desc = L["Modify "] .. talentName,
							min = 0,
							max = maxRank,
							step = 1,
							order = i + (t-1) * 35,
							get =  function() 
									if self.talents[talentName] then
										return self.talents[talentName]
									else
										return 0
									end
								end,
							set = 	function(v)
									local value = math_floor(v+0.5)
									if value == 0 then
										self.talents[talentName] = nil
									else
										self.talents[talentName] = value
									end
									self:UpdateTalents(true)
								end,
					}
				end
				if (currRank ~= 0 or all) and talentInfo[talentName] then
					talents[talentName] = all and maxRank or currRank
				end	    	    
			end
		end
	end
	--Load talents into appropriate spell data
	for _, spell in pairs( spellInfo ) do
		TalentIterate( spell["Name"], spell[0], spell["Talents"] )
		if spell["Secondary"] then
			TalentIterate( spell["Secondary"]["Name"], spell["Secondary"][0], spell["Secondary"]["Talents"] )
		end
	end	
	if not self:IsEventScheduled("DrD_FullUpdate") then
		self:UpdateAB()
	end
end

TalentIterate = function( name, baseSpell, endtable )
	if not baseSpell or type(baseSpell) == "function" then return end
	DrD_ClearTable( endtable )
	local melee = baseSpell.Melee
	local school = baseSpell.School
	local s1,s2,s3
	if type(school) == "table" then
		s1 = school[1]
		s2 = school[2]
		s3 = school[3]
	else
		s1 = school or "Physical"
	end
	for talentName, talentRank in pairs( talents ) do
		local talentTable = talentInfo[talentName]
		for i=1,#talentTable do
			local talent = talentTable[i]
			if (melee and talent.Melee or talent.Caster and not melee or (not talent.Melee and not talent.Caster)) and not (talent.Specials and (baseSpell.AutoAttack or baseSpell.AutoShot)) then
				local spells = talent.Spells
				if DrD_MatchData(spells,name,s2,s3) or not baseSpell.NoSchoolTalents and DrD_MatchData(spells, "All", s1) then
					local number = #endtable + 1
					local modtype = talent.ModType
					local multiply = talent.Multiply
					endtable["Multiply"..number] = multiply
					endtable[number] = (type(talent.Effect) == "table") and talent.Effect[talentRank] or talent.Effect * talentRank
					if not modtype then
						endtable[("ModType"..number)] = multiply and "dmgM" or "dmgM_Add"
					elseif modtype == "SpellDamage" then
						endtable[("ModType"..number)] = multiply and "spellDmgM" or "spellDmgM_Add"
					else
						endtable[("ModType"..number)] = modtype
					end
					if talent.Mod then
						local mod = talent.Mod
						if DrD_MatchData(mod[1], name, "All", s1, s2, s3) then
							mod[6] = talentRank
							endtable["Mod"..number] = mod
						end
					end
				end
			end
		end
	end
end

local debug_flag

local SetItems = {}
function DrDamage:GetSetAmount(set)
	if debug_flag then
		return 8
	end
	if not updateSetItems and SetItems[set] then
		return SetItems[set]
	end
	
	if updateSetItems then
		DrD_ClearTable( SetItems )
		updateSetItems = false
	end
	
	local amount = 0
	local setData = self.SetBonuses[set]
	if setData then
		for _, itemID in ipairs(setData) do
			if IsEquippedItem(itemID) then
				amount = amount + 1
			end
		end
	end
	SetItems[set] = amount
	return amount
end

local Glyphs = {}
function DrDamage:HasGlyph(glyph)
	return debug_flag or Glyphs[glyph]
end

function DrDamage:UpdateGlyphs()
	DrD_ClearTable(Glyphs)
	for i=1,6 do
		local _, _, id = GetGlyphSocketInfo(i)
		if id then Glyphs[id] = true end
	end
end

function DrDamage:SetSpell( frame, slot )
	if settings.NeverTooltip or (settings.CombatTooltip and InCombatLockdown() or not settings.CombatTooltip) and (not settings.AlwaysTooltip and not (settings.AltTooltip and IsAltKeyDown()) and not (settings.CtrlTooltip and IsControlKeyDown()) and not (settings.ShiftTooltip and IsShiftKeyDown())) then
		return
	end
	
	local name, rank = GetSpellName(slot,BOOKTYPE_SPELL)
	
	if spellInfo[name] then
		local baseSpell = spellInfo[name][0]
		if type(baseSpell) == "function" then baseSpell = baseSpell() end

		if baseSpell and not baseSpell.NoTooltip then	
			if playerCaster or playerHybrid and not baseSpell.Melee then
				self:CasterTooltip( frame, name, rank )
			elseif playerMelee or playerHybrid and baseSpell.Melee then
				self:MeleeTooltip( frame, name, rank )
			end
		end
	end
end

function DrDamage:SetAction( frame, slot )
	self.hooks[frame].SetAction(frame, slot)
	
	if settings.NeverTooltip or (settings.CombatTooltip and InCombatLockdown() or not settings.CombatTooltip) and (not settings.AlwaysTooltip and not (settings.AltTooltip and IsAltKeyDown()) and not (settings.CtrlTooltip and IsControlKeyDown()) and not (settings.ShiftTooltip and IsShiftKeyDown())) then
		return
	end
	
	local gtype, pid = GetActionInfo(slot)
	local name, rank
		
	if (gtype == "spell" and pid ~= 0) then
		name, rank = GetSpellName(pid, gtype)
	elseif gtype == "macro" then
		name, rank = GetMacroSpell(pid)
	end

	if name and spellInfo[name] then
		local baseSpell = spellInfo[name][0]
		if type(baseSpell) == "function" then baseSpell = baseSpell() end

		if baseSpell and not baseSpell.NoTooltip then
			rank = rank or select(2,GetSpellInfo(name))
			if playerCaster or playerHybrid and not baseSpell.Melee then
				self:CasterTooltip( frame, name, rank )
			elseif playerMelee or playerHybrid and baseSpell.Melee then
				self:MeleeTooltip( frame, name, rank )
			end
		end
	end
end

local DrD_ProcessButton, DrD_CreateText, DrD_SpecialText
function DrDamage:UpdateAB(spell, uid)

	self:TriggerEvent("DrDamage_Update", spell, uid)
	self:CancelSpellUpdate()

	--	Used for debugging updates.
	--[[
	if spell then
		self:Print( "Update AB: Spell" )
		
		if type( spell ) == "table" then
			for _, gar in ipairs( spell ) do
				self:Print( "Updating: " .. gar )
			end
		else
			self:Print( "Updating: " .. spell )
		end
	elseif uid then
		self:Print( "Update AB: ID" )
		
	else
		self:Print( "Update AB: All" )
	end
	--]]
	
	if ABfunc then
		ABfunc()
	end

	if not ABdefault then
		if settings.BlizzardAB then 
			ABdefault = true 
		elseif ABstop then
			return
		end
	elseif not settings.BlizzardAB and ABtable then
		ABdisable = true
		ABdefault = false
	end
	
	if ABtable then
		for name, func in pairs(ABtable) do
			DrD_ProcessButton(_G[name], func, spell, uid)
		end	
	end
	if ABdefault or ABdisable then
		for i, name in ipairs(ABobjects) do
			DrD_ProcessButton(name, nil, spell, uid)
		end
	end
					
	self.visualChange = nil
	ABdisable = false
end

DrD_ProcessButton = function(button, func, spell, uid)
	if not button then return end 
	if not settings.ABText or ABdisable then 
		if button.drd then button.drd:Hide() end		
		return 
	end
	if button:IsVisible() then --or ABhidden then
		if not spell and not uid then
			local frame = button.drd
			if frame then frame:Hide() end
		end	
		local id, name, rank
		
		if func then
			id, name, rank = func(button)
		else
			id = ActionButton_GetPagedID(button)
		end
		if id and (not HasAction(id) or uid and uid ~= id) then return end
		DrDamage:CheckAction(button, spell, id, name, rank)
	end
end

function DrDamage:CheckAction(button, spell, id, name, rank)
	local gtype, pid
	if id then
		gtype, pid = GetActionInfo(id)

		if gtype == "spell" and (pid > 0) then
			name, rank = GetSpellName(pid, gtype)
		elseif gtype == "macro" then
			name, rank = GetMacroSpell(pid)
		end
	end
	if name then
		if spell and not DrD_MatchData(spell, name) then
			return
		end
		if rank and tonumber(rank) and GetSpellInfo(name) then
			rank = string_gsub(select(2,GetSpellInfo(name)),"%d+", rank)
		end
		if self.ClassSpecials[name] then
			return DrD_SpecialText(self.ClassSpecials[name], rank, button)
		end
		if spellInfo[name] then
			local baseSpell = settings.SwapCalc and spellInfo[name]["Secondary"] and spellInfo[name]["Secondary"][0] or spellInfo[name][0]
			if type(baseSpell) == "function" then baseSpell = baseSpell() end

			if baseSpell then
				if gtype == "macro" and button then
					local macroText = _G[button:GetName().."Name"]
					if macroText then macroText:Hide() end
				end			

				local text, healingSpell

				if playerCaster or playerHybrid and not baseSpell.Melee then
					healingSpell = baseSpell.Healing or DrD_MatchData(baseSpell.School, "Healing")

					if (settings.CastsLeft and healingSpell or settings.CastsLeftDmg and not healingSpell) and not baseSpell.NoDPM then
						local _, _, _, manaCost = GetSpellInfo(name,rank)

						if manaCost and manaCost > 0 then
							if UnitPowerType("player") == 0 then
								text = math_floor(UnitMana("player") / manaCost)
							else
								text = math_floor(self.lastMana / manaCost)	
							end
						else
							text = "\226\136\158"
						end
					else
						text = self:CasterCalc(name, rank)								
					end
				elseif playerMelee or playerHybrid and baseSpell.Melee then
					text = self:MeleeCalc(name, rank)
				end

				if type(text) == "number" then
					text = math_floor(text + 0.5)
				end

				if button then
					DrD_CreateText(text, button, healingSpell)
				elseif text then
					if healingSpell then
						return "|cff".. string_format("%02x%02x%02x", settings.FontColorHeal.r * 255, settings.FontColorHeal.g * 255, settings.FontColorHeal.b * 255).. text .. "|r"
					else
						return "|cff".. string_format("%02x%02x%02x", settings.FontColorDmg.r * 255, settings.FontColorDmg.g * 255, settings.FontColorDmg.b * 255).. text .. "|r"
					end
				end
			end
		end
	end
end

DrD_CreateText = function(text, button, healing, r, g, b )
	local drd = button.drd

	if not r and (DrDamage.visualChange or not drd) then
		local color = healing and settings.FontColorHeal or settings.FontColorDmg
		r,g,b = color.r, color.g, color.b
	end		
	
	if drd then
		if DrDamage.visualChange then
			drd:SetFont(DrD_Font, settings.FontSize, settings.FontEffect)
			drd:SetPoint("BOTTOMLEFT", button, "BOTTOMLEFT", -10 + settings.FontXPosition, settings.FontYPosition + 5)
			drd:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 10 + settings.FontXPosition, settings.FontYPosition + 5)
			drd:SetTextColor(r,g,b)	
		end
		drd:SetText(text)
		drd:Show()
	else
		drd = button:CreateFontString("drd", "OVERLAY")
		button.drd = drd
		drd:SetPoint("BOTTOMLEFT", button, "BOTTOMLEFT", -10 + settings.FontXPosition, settings.FontYPosition + 5)
		drd:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 10 + settings.FontXPosition, settings.FontYPosition + 5)
		drd:SetFont(DrD_Font, settings.FontSize, settings.FontEffect)
		drd:SetJustifyH("CENTER")
		drd:SetTextColor(r,g,b)
		drd:SetText(text)
		drd:Show()
	end
end

DrD_SpecialText = function( func, rank, button )
	local text, r, g, b = func(rank)
	if type(text) == "number" then text = math_floor(text + 0.5) end
	if text then
		if button then 
			DrD_CreateText(text, button, false, r, g, b)
		else
			if r then
				return "|cff".. string_format("%02x%02x%02x", r * 255, g * 255, b * 255) .. text .. "|r"
			else
				return "|cff".. string_format("%02x%02x%02x", settings.FontColorDmg.r * 255, settings.FontColorDmg.g * 255, settings.FontColorDmg.b * 255) .. text .. "|r"
			end
		end
	end
end

--Credits to the author of RatingBuster (Whitetooth) for the formula!
local ratingTypes = { ["Hit"] = 8, ["Crit"] = 14, ["MeleeHit"] = 10, ["SpellHaste"] = 10 }
function DrDamage:GetRating( rType, convertR, full )
	local playerLevel = UnitLevel("player")
	local base = ratingTypes[rType]
	local rating, value

	if playerLevel > 70 then
		rating = base * (82/52) * ((131/63)^((playerLevel - 70)/10))
	elseif playerLevel > 60 then
		rating = base * 82 / (262 - 3 * playerLevel)
	elseif playerLevel > 10 then
		rating = base * ((playerLevel - 8) / 52)
	elseif 	playerLevel <= 10 then
		rating = base / 26
	end

	value = convertR and convertR/rating or rating
	value = full and value or DrD_Round(value,2)
	return value
end

function DrDamage:GetLevels()
	local targetLevel = UnitLevel("target")
	local playerLevel = UnitLevel("player")
	
	if (UnitClassification("target") == "worldboss") then
		targetLevel = playerLevel + 3
	elseif targetLevel == -1 then
		targetLevel = playerLevel + 10
	end
	
	return math_min(10, targetLevel - playerLevel), playerLevel, targetLevel
end

local oldRangedItem
function DrDamage:CheckRelicSlot()
	local newRangedItem
	if GetInventoryItemLink("player", GetInventorySlotInfo("RangedSlot")) then
		newRangedItem = GetItemInfo(GetInventoryItemLink("player", GetInventorySlotInfo("RangedSlot")))
	end
	if newRangedItem ~= oldRangedItem then
		oldRangedItem = newRangedItem
		return true
	end
end

local WeaponBuffScan = GetTime()
local WeaponBuff, WeaponBuffRank 
function DrDamage:GetWeaponBuff(off)
	local mh, _, _, oh = GetWeaponEnchantInfo()
	local name, rank, buff
	
	if not off and mh then
		if GetTime() > WeaponBuffScan then
			WeaponBuffScan = GetTime() + 2
			GT:SetInventoryItem("player", GetInventorySlotInfo("MainHandSlot"))
			_, _, buff = GT:Find("^([^%(]+) %(%d+ [^%)]+%)$", nil, nil, false, true)
			if buff then
				name, rank = string_match(buff,"^(.*) (%d+)$")
			end
			WeaponBuff, WeaponBuffRank = name or buff, rank
		end
		return WeaponBuff, WeaponBuffRank
	elseif off and oh then
		GT:SetInventoryItem("player", GetInventorySlotInfo("SecondaryHandSlot"))
		_, _, name, rank = GT:Find("^([^%(]+) %(%d+ [^%)]+%)$", nil, nil, false, true)
		if buff then
			name, rank = string_match(buff,"^(.*) (%d+)$")
		end
	end
	return name or buff, rank
end

function DrDamage:Calc(name, rank, tooltip, modify, debug)
	if not spellInfo or not name then return end
	if not spellInfo[name] then return end
	local baseSpell = settings.SwapCalc and spellInfo[name]["Secondary"] and spellInfo[name]["Secondary"][0] or spellInfo[name][0]
	if type(baseSpell) == "function" then baseSpell = baseSpell() end
	if baseSpell then
		if playerCaster or playerHybrid and not baseSpell.Melee then
			return self:CasterCalc(name, rank, tooltip, modify, debug)								
		elseif playerMelee or playerHybrid and baseSpell.Melee then
			return self:MeleeCalc(name, rank, tooltip, modify, debug)
		end
	end
end

function DrDamage:TestAll()
	debug_flag = true
	local spells, ranks = 0, 0
	self:UpdateTalents(false,true)
	for k, v in pairs( spellInfo ) do
		local baseSpell = spellInfo[k][0]
		for i=1,#spellInfo[k] do
			if playerCaster or playerHybrid and not baseSpell.Melee then
				self:CasterCalc(k, i, true, nil, true)								
			elseif playerMelee or playerHybrid and baseSpell.Melee then
				self:MeleeCalc(k, i, true, nil, true)
			end
			ranks = ranks + 1
		end
		spells = spells + 1
	end
	debug_flag = false
	self:Print( "Tested: " .. ranks .. " ranks of " .. spells .. " spells.")
end

--[[ Used for harvesting base value increases:
function DrDamage:FindIDs(name)
	local stid = {}
	for i=1,100000 do
		local sname, srank = GetSpellInfo(i)
		srank = srank and string_match(srank,"%d+")
		if  name == sname and srank then
			stid[i] = true
		end
	end
	return stid
end

function DrDamage:HarvestValues(id)
	GT:SetHyperlink(GetSpellLink(id))
	local name, rank = GetSpellInfo(id)
	rank = rank and tonumber(string_match(rank,"%d+")) --or 1
	local min, max = string_match(GT:GetLine(GT:NumLines()),"(%d+) to (%d+)")
	if not min then 
		min = string_match(GT:GetLine(GT:NumLines()),"causing (%d+) ")
		min = min or string_match(GT:GetLine(GT:NumLines()),"Causes (%d+) ")
		min = min or string_match(GT:GetLine(GT:NumLines()),"doing (%d+) ")
		min = min or string_match(GT:GetLine(GT:NumLines()),"absorbing (%d+) ")
		min = min or string_match(GT:GetLine(GT:NumLines()),"Absorbs (%d+) ")
		min = min or string_match(GT:GetLine(GT:NumLines()),"deals (%d+) ")
		--min = min or string_match(GT:GetLine(GT:NumLines()),"for (%d+) ")
		--min = min or string_match(GT:GetLine(GT:NumLines()),"(%d+)%.")
		max = min
	end
	self:Print(name, rank, min, max)
	return name, rank, min, max
end

function DrDamage:Wowhead()
	--local stuff = ""
	--local table = {}
	--for k in string.gmatch(stuff,"spell=(%d+)") do
	--	table[k] = true
	--end
	--return table
	return { [47540] = true, [53005] = true, [53006] = true, [53007] = true }
end

local haxtable = {}
function DrDamage:Hax(value,force,remove,secondary)
	if value then
		local error
		if not remove then remove = 0 end
		DrD_ClearTable( haxtable )
		local entries = 0
		for k in pairs((type(value) == "table") and value or DrDamage:FindIDs(value)) do
			entries = entries + 1
			local name, rank, min, max = DrDamage:HarvestValues(k)
			local data = secondary and DrDamage.dump[name] and DrDamage.dump[name]["Secondary"] or DrDamage.dump[name]
			ChatFrame1:AddMessage(GetSpellLink(k))
			if name and rank and data and data[rank] and min and max then
				local oldmin, oldmax, spellLevel = data[rank][1], data[rank][2]
				if haxtable[rank] and ((haxtable[rank][1] ~= min) or (haxtable[rank][2] ~= max)) then
					self:Print("WARNING: Rank " .. rank .. " already exists in table with different values.")
					error = true
				else
					haxtable[rank] = { min, max, oldmin, oldmax, data[rank] }
				end
			end
		end
		self:Print("NOTE: Spell ID entries: " .. entries)
		if #haxtable > 0 then
			for i=1,#haxtable do
				local val = haxtable[i]
				if val then
					self:Print( "Rank " .. i .. ": " .. val[1] .. "-" .. val[2] .. " from " .. val[3] .. "-" .. val[4] )
				else
					self:Print( "Rank " .. i .. ": Error")
					error = true
				end
					
			end
			if not error or force then
				self:Print("Difference table:")
				for i=1,#haxtable do
					local val = haxtable[i]
					local rankstring = "[" .. i .. "]" .. " = { " .. val[3] .. ", " .. val[4] .. ", " .. (val[1] - val[3] - remove) .. ", " .. (val[2] - val[4] - remove)
					for k, v in pairs(val[5]) do
						if type(k) ~= "number" then
							rankstring = rankstring .. ", " .. k .. " = " .. v
						end
					end
					rankstring = rankstring .. ", },"
					ChatFrame1:AddMessage( rankstring ) 
				end
			end
		end
	end
end

DrDamage.dump = { }
--]]
